﻿using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using RhythmBullet.Zer01HD.Game.ContentResolvers;
using RhythmBullet.Zer01HD.Game.Preferences;
using RhythmBullet.Zer01HD.Game.Screens;
using RhythmBullet.Zer01HD.Utilities;
using RhythmBullet.Zer01HD.Utilities.ContentSystem;
using RhythmBullet.Zer01HD.Utilities.DataTypes;
using RhythmBullet.Zer01HD.Utilities.Input;
using RhythmBullet.Zer01HD.Utilities.UI.ScreenSystem;
using System;
using System.Diagnostics;

namespace RhythmBullet
{
    /// <summary>
    /// This is the main type for your game.
    /// </summary>
    public class RhythmBulletGame : Game
    {
        public const int WORLD_WIDTH = 4;
        public const int WORLD_HEIGHT = 3;
        public static int pixels_per_WU;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        public readonly ContentSystem Assets;
        readonly ResolutionContentResolver resolutionContentResolver;
        public PreferencesManager preferencesManager;
        private Screen previousScreen;
        private RenderTarget2D previousScreenRenderTarget;
        private Screen currentScreen;
        private bool resizing;
        public Screen Screen
        {
            get
            {
                return currentScreen;
            }
            set
            {
                previousScreen = currentScreen;
                currentScreen = value;
                if (!Screen.Initiated)
                {
                    Screen.Initiate(graphics.GraphicsDevice, new Rectangle(0, 0, graphics.PreferredBackBufferWidth, graphics.PreferredBackBufferHeight));
                }
                if (previousScreen != null && previousScreenRenderTarget == null && previousScreen.UseRenderTargetForExitTransition)
                {
                    previousScreenRenderTarget = new RenderTarget2D(graphics.GraphicsDevice, graphics.PreferredBackBufferWidth, graphics.PreferredBackBufferHeight);
                }
                graphics.GraphicsDevice.SetRenderTarget(previousScreenRenderTarget);
                graphics.GraphicsDevice.Clear(Color.Black);

                Debug.WriteLine("Showing " + value.GetType().Name);
                Debug.WriteLine("Previous screen is " + previousScreen?.GetType().Name);
                Screen.Show();
            }
        }

        public RhythmBulletGame()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
            Assets = new ContentSystem(Content);
            resolutionContentResolver = new ResolutionContentResolver();
            FontContentResolver fcr = new FontContentResolver(resolutionContentResolver);
            Assets.contentPathModifier.Add(typeof(Texture2D), resolutionContentResolver);
            Assets.contentPathModifier.Add(typeof(SpriteFont), fcr);
            preferencesManager = new PreferencesManager(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) + "/RhythmBullet",
                new General(),
                new Controls());
            preferencesManager.Load();
            Resolution resolution = preferencesManager.GetPreferences<General>().Resolution;
            graphics.PreferredBackBufferWidth = resolution.Width;
            graphics.PreferredBackBufferHeight = resolution.Height;
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            Resolution resolution = preferencesManager.GetPreferences<General>().Resolution;
            resolutionContentResolver.Width = resolution.Width;
            resolutionContentResolver.Height = resolution.Height;
            Debug.WriteLine("Initial setup complete.");
            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);
            QueueContent();
            Screen = new LoadingScreen(Content.Load<Texture2D>("RhythmBullet"), 0.7f);
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// game-specific content.
        /// </summary>
        protected override void UnloadContent()
        {
            Assets.UnloadAll();
            previousScreenRenderTarget?.Dispose();
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.   
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            if (!Assets.Done)
            {
                Assets.Update();
            }
            else if (resizing)
            {
                resizing = false;
                PostResize();
            }
            switch (Screen.State)
            {
                case ScreenState.EnterTransition:
                    if (previousScreen != null && previousScreen.UseRenderTargetForExitTransition)
                    {
                        previousScreen.UpdateTransition(gameTime.ElapsedGameTime.TotalSeconds, Assets.Done);
                    }
                    Screen.UpdateTransition(gameTime.ElapsedGameTime.TotalSeconds, Assets.Done);
                    break;
                case ScreenState.ExitTransition:
                    if (Screen.UseRenderTargetForExitTransition || Screen.UpdateTransition(gameTime.ElapsedGameTime.TotalSeconds, Assets.Done))
                    {
                        if (Screen.NextScreen != null)
                        {
                            Debug.WriteLine("Changing to the next given screen.");
                            Screen = Screen.NextScreen;
                        }
                        else
                        {
                            Debug.WriteLine("Changing to default screen.");
                            Screen = previousScreen != null ? previousScreen : new MainScreen(Assets);
                        }
                    }
                    break;
                case ScreenState.Normal:
                    Screen.Update(gameTime);
                    break;
            }

            InputUtilities.Update();
            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Screen.BackgroundColor);
            if (Screen.State == ScreenState.EnterTransition && previousScreen != null && previousScreen.UseRenderTargetForExitTransition)
            {
                graphics.GraphicsDevice.SetRenderTarget(previousScreenRenderTarget);
                GraphicsDevice.Clear(previousScreen.BackgroundColor);
                spriteBatch.Begin();
                previousScreen.Draw(spriteBatch);
                spriteBatch.End();
                graphics.GraphicsDevice.SetRenderTarget(null);
                Screen.UpdatePreviousScreenFrame(previousScreenRenderTarget);
            }
            spriteBatch.Begin();
            Screen.Draw(spriteBatch);
            spriteBatch.End();

            base.Draw(gameTime);
        }

        public void Resize(int width, int height)
        {
            pixels_per_WU = height / WORLD_HEIGHT;
            resolutionContentResolver.Width = width;
            resolutionContentResolver.Height = height;
            graphics.PreferredBackBufferWidth = width;
            graphics.PreferredBackBufferHeight = height;
            resizing = true;
            Assets.UnloadAll();
            QueueContent();
            Screen.AssetLoadStateChange(true);
            Screen = new LoadingScreen(Content.Load<Texture2D>("loading_ring"), 0.3f, true);
            previousScreenRenderTarget.Dispose();
            previousScreenRenderTarget = null;
        }

        private void PostResize()
        {
            graphics.ApplyChanges();
            Screen.AssetLoadStateChange(false);
        }

        private void QueueContent()
        {
            Assets.Queue<Texture2D>("title");
            Assets.Queue<Texture2D>("default_cover", false);
            Assets.Queue<Texture2D>("backgrounds/mainBG");
        }
    }
}
